@@ -13,7 +13,6 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"reflect"
 	"strings"
 	"testing"
 )
@@ -39,6 +38,7 @@ func TestStreamWriteBlockedByOutputBuffer(t *testing.T) {
 		if n != writeBufferSize || err != context.Canceled {
 			t.Fatalf("s.WriteContext() = %v, %v; want %v, context.Canceled", n, err, writeBufferSize)
 		}
+		s.Flush()
 		tc.wantFrame("first write buffer of data sent",
 			packetType1RTT, debugFrameStream{
 				id:   s.id,
@@ -48,7 +48,9 @@ func TestStreamWriteBlockedByOutputBuffer(t *testing.T) {
 
 		// Blocking write, which must wait for buffer space.
 		w := runAsync(tc, func(ctx context.Context) (int, error) {
-			return s.WriteContext(ctx, want[writeBufferSize:])
+			n, err := s.WriteContext(ctx, want[writeBufferSize:])
+			s.Flush()
+			return n, err
 		})
 		tc.wantIdle("write buffer is full, no more data can be sent")
 
@@ -171,6 +173,7 @@ func TestStreamIgnoresMaxStreamDataReduction(t *testing.T) {
 			t.Fatal(err)
 		}
 		s.WriteContext(ctx, want[:1])
+		s.Flush()
 		tc.wantFrame("sent data (1 byte) fits within flow control limit",
 			packetType1RTT, debugFrameStream{
 				id:   s.id,
@@ -724,7 +727,7 @@ func testStreamSendFrameInvalidState(t *testing.T, f func(sid streamID) debugFra
 		if err != nil {
 			t.Fatal(err)
 		}
-		s.Write(nil) // open the stream
+		s.Flush() // open the stream
 		tc.wantFrame("new stream is opened",
 			packetType1RTT, debugFrameStream{
 				id:   sid,
@@ -848,7 +851,7 @@ func TestStreamOffsetTooLarge(t *testing.T) {
 	got, _ := tc.readFrame()
 	want1 := debugFrameConnectionCloseTransport{code: errFrameEncoding}
 	want2 := debugFrameConnectionCloseTransport{code: errFlowControl}
-	if !reflect.DeepEqual(got, want1) && !reflect.DeepEqual(got, want2) {
+	if !frameEqual(got, want1) && !frameEqual(got, want2) {
 		t.Fatalf("STREAM offset exceeds 2^62-1\ngot:  %v\nwant: %v\n  or: %v", got, want1, want2)
 	}
 }
@@ -969,7 +972,9 @@ func TestStreamWriteMoreThanOnePacketOfData(t *testing.T) {
 	want := make([]byte, 4096)
 	rand.Read(want) // doesn't need to be crypto/rand, but non-deprecated and harmless
 	w := runAsync(tc, func(ctx context.Context) (int, error) {
-		return s.WriteContext(ctx, want)
+		n, err := s.WriteContext(ctx, want)
+		s.Flush()
+		return n, err
 	})
 	got := make([]byte, 0, len(want))
 	for {
@@ -999,6 +1004,7 @@ func TestStreamCloseWaitsForAcks(t *testing.T) {
 	tc, s := newTestConnAndLocalStream(t, serverSide, uniStream, permissiveTransportParameters)
 	data := make([]byte, 100)
 	s.WriteContext(ctx, data)
+	s.Flush()
 	tc.wantFrame("conn sends data for the stream",
 		packetType1RTT, debugFrameStream{
 			id:   s.id,
@@ -1041,11 +1047,13 @@ func TestStreamCloseUnblocked(t *testing.T) {
 	for _, test := range []struct {
 		name    string
 		unblock func(tc *testConn, s *Stream)
+		success bool
 	}{{
 		name: "data received",
 		unblock: func(tc *testConn, s *Stream) {
 			tc.writeAckForAll()
 		},
+		success: true,
 	}, {
 		name: "stop sending received",
 		unblock: func(tc *testConn, s *Stream) {
@@ -1065,6 +1073,7 @@ func TestStreamCloseUnblocked(t *testing.T) {
 			tc, s := newTestConnAndLocalStream(t, serverSide, uniStream, permissiveTransportParameters)
 			data := make([]byte, 100)
 			s.WriteContext(ctx, data)
+			s.Flush()
 			tc.wantFrame("conn sends data for the stream",
 				packetType1RTT, debugFrameStream{
 					id:   s.id,
@@ -1087,7 +1096,13 @@ func TestStreamCloseUnblocked(t *testing.T) {
 				t.Fatalf("s.CloseContext() = %v, want it to block waiting for acks", err)
 			}
 			test.unblock(tc, s)
-			if _, err := closing.result(); err != nil {
+			_, err := closing.result()
+			switch {
+			case err == errNotDone:
+				t.Fatalf("s.CloseContext() still blocking; want it to have returned")
+			case err == nil && !test.success:
+				t.Fatalf("s.CloseContext() = nil, want error")
+			case err != nil && test.success:
 				t.Fatalf("s.CloseContext() = %v, want nil (all data acked)", err)
 			}
 		})
@@ -1229,6 +1244,7 @@ func TestStreamPeerStopSendingForActiveStream(t *testing.T) {
 		tc, s := newTestConnAndLocalStream(t, serverSide, styp, permissiveTransportParameters)
 		for i := 0; i < 4; i++ {
 			s.Write([]byte{byte(i)})
+			s.Flush()
 			tc.wantFrame("write sends a STREAM frame to peer",
 				packetType1RTT, debugFrameStream{
 					id:   s.id,
@@ -1272,6 +1288,99 @@ func TestStreamReceiveDataBlocked(t *testing.T) {
 	tc.wantIdle("no response to STREAM_DATA_BLOCKED and DATA_BLOCKED")
 }
 
+func TestStreamFlushExplicit(t *testing.T) {
+	testStreamTypes(t, "", func(t *testing.T, styp streamType) {
+		tc, s := newTestConnAndLocalStream(t, clientSide, styp, permissiveTransportParameters)
+		want := []byte{0, 1, 2, 3}
+		n, err := s.Write(want)
+		if n != len(want) || err != nil {
+			t.Fatalf("s.Write() = %v, %v; want %v, nil", n, err, len(want))
+		}
+		tc.wantIdle("unflushed data is not sent")
+		s.Flush()
+		tc.wantFrame("data is sent after flush",
+			packetType1RTT, debugFrameStream{
+				id:   s.id,
+				data: want,
+			})
+	})
+}
+
+func TestStreamFlushImplicitExact(t *testing.T) {
+	testStreamTypes(t, "", func(t *testing.T, styp streamType) {
+		const writeBufferSize = 4
+		tc, s := newTestConnAndLocalStream(t, clientSide, styp,
+			permissiveTransportParameters,
+			func(c *Config) {
+				c.MaxStreamWriteBufferSize = writeBufferSize
+			})
+		want := []byte{0, 1, 2, 3, 4, 5, 6}
+
+		// This write doesn't quite fill the output buffer.
+		n, err := s.Write(want[:3])
+		if n != 3 || err != nil {
+			t.Fatalf("s.Write() = %v, %v; want %v, nil", n, err, len(want))
+		}
+		tc.wantIdle("unflushed data is not sent")
+
+		// This write fills the output buffer exactly.
+		n, err = s.Write(want[3:4])
+		if n != 1 || err != nil {
+			t.Fatalf("s.Write() = %v, %v; want %v, nil", n, err, len(want))
+		}
+		tc.wantFrame("data is sent after write buffer fills",
+			packetType1RTT, debugFrameStream{
+				id:   s.id,
+				data: want[0:4],
+			})
+
+	})
+}
+
+func TestStreamFlushImplicitLargerThanBuffer(t *testing.T) {
+	testStreamTypes(t, "", func(t *testing.T, styp streamType) {
+		const writeBufferSize = 4
+		tc, s := newTestConnAndLocalStream(t, clientSide, styp,
+			permissiveTransportParameters,
+			func(c *Config) {
+				c.MaxStreamWriteBufferSize = writeBufferSize
+			})
+		want := []byte{0, 1, 2, 3, 4, 5, 6, 7, 8, 9}
+
+		w := runAsync(tc, func(ctx context.Context) (int, error) {
+			n, err := s.WriteContext(ctx, want)
+			return n, err
+		})
+
+		tc.wantFrame("data is sent after write buffer fills",
+			packetType1RTT, debugFrameStream{
+				id:   s.id,
+				data: want[0:4],
+			})
+		tc.writeAckForAll()
+		tc.wantFrame("ack permits sending more data",
+			packetType1RTT, debugFrameStream{
+				id:   s.id,
+				off:  4,
+				data: want[4:8],
+			})
+		tc.writeAckForAll()
+
+		tc.wantIdle("write buffer is not full")
+		if n, err := w.result(); n != len(want) || err != nil {
+			t.Fatalf("Write() = %v, %v; want %v, nil", n, err, len(want))
+		}
+
+		s.Flush()
+		tc.wantFrame("flush sends last buffer of data",
+			packetType1RTT, debugFrameStream{
+				id:   s.id,
+				off:  8,
+				data: want[8:],
+			})
+	})
+}
+
 type streamSide string
 
 const (
@@ -1289,31 +1398,41 @@ func newTestConnAndStream(t *testing.T, side connSide, sside streamSide, styp st
 
 func newTestConnAndLocalStream(t *testing.T, side connSide, styp streamType, opts ...any) (*testConn, *Stream) {
 	t.Helper()
-	ctx := canceledContext()
 	tc := newTestConn(t, side, opts...)
 	tc.handshake()
 	tc.ignoreFrame(frameTypeAck)
+	return tc, newLocalStream(t, tc, styp)
+}
+
+func newLocalStream(t *testing.T, tc *testConn, styp streamType) *Stream {
+	t.Helper()
+	ctx := canceledContext()
 	s, err := tc.conn.newLocalStream(ctx, styp)
 	if err != nil {
 		t.Fatalf("conn.newLocalStream(%v) = %v", styp, err)
 	}
-	return tc, s
+	return s
 }
 
 func newTestConnAndRemoteStream(t *testing.T, side connSide, styp streamType, opts ...any) (*testConn, *Stream) {
 	t.Helper()
-	ctx := canceledContext()
 	tc := newTestConn(t, side, opts...)
 	tc.handshake()
 	tc.ignoreFrame(frameTypeAck)
+	return tc, newRemoteStream(t, tc, styp)
+}
+
+func newRemoteStream(t *testing.T, tc *testConn, styp streamType) *Stream {
+	t.Helper()
+	ctx := canceledContext()
 	tc.writeFrames(packetType1RTT, debugFrameStream{
-		id: newStreamID(side.peer(), styp, 0),
+		id: newStreamID(tc.conn.side.peer(), styp, 0),
 	})
 	s, err := tc.conn.AcceptStream(ctx)
 	if err != nil {
 		t.Fatalf("conn.AcceptStream() = %v", err)
 	}
-	return tc, s
+	return s
 }
 
 // permissiveTransportParameters may be passed as an option to newTestConn.
